#ifndef ATTACK_H
#define ATTACK_H

#include "Global.h"
#include "ObjectIlk.h"

enum shEnergyType {
    kNoEnergy,
    kSlashing,
    kPiercing,
    kConcussive,

    kBlinding,
//    kBrainDraining,  /* neuralizer, brain bug */
    kBrainExtracting,/* mi-go */
    kBugging,
    kBurning,
    kChoking,
    kConfusing,
    kCorrosive,      /* acid */
    kCreditDraining, /* creeping credits :-) */
    kDigging,
    kDisarming,
    kDisintegrating, /* antimatter */
    kElectrical,
    kFaceHugging,
    kFlare,
    kForce,
    kLaser,
    kFreezing,
    kHealing,
    kHosing,         /* ftp daemon breathe traffic */
    kMagnetic,
    kMesmerizing,    /* hypnotic gaze */
    kParalyzing,
    kPlaguing,       /* defiler vomit */
    kPoisonous,
    kRadiological,
    kRestoring,
    kPsychic,
    kSickening,      /* viruses */
    kStunning,
    kTimeWarping,
    kTransporting,
    kViolating,
    kWebbing,
    kMaxEnergyType
};

#define IS_SPECIAL_ENERGY(_e) ((_e) > kConcussive)

enum shAttackFlags {
    kMelee =           0x1,  /* mop, chainsaw */
    kMissile =         0x2,  /* dagger, grenade */
    kAimed =           0x4,  /* blaster, chaingun */
    kTouchAttack =     0x10, /* AC is 10 + bonus from agility */
    kSecondaryAttack = 0x20  /* half bonus from strength */
};

struct shAttack
{
    enum Type {        /* primarily used for descriptive purposes */
        kNoAttack,
        kAcidBath,     /* acid trap */
        kAcidSplash,   /* acid blood */
        kAnalProbe,
        kAttach,       /* e.g. attaching a restraining bolt */
        kBite,
        kBlast,        /* explosion */
        kBullet,
        kBolt,         /* energy bolt, e.g. pea shooter, blaster bolt */
        kBreatheBugs,
        kBreatheFire,
        kBreatheTime,
        kBreatheTraffic,
        kBreatheViruses,
        kClaw,
        kClub,
        kChoke,
        kCombi,        /* combi-stick */
        kCook,
        kCrush,
        kCut,
        kDisc,         /* Smart-Disc */
        kDisintegrationRay,
        kExplode,
        kExtractBrain, /* mi-go */
        kFaceHug,      /* facehugger */
        kFlash,        /* radiation grenade */
        kFreezeRay,
        kGammaRay,
        kGaussRay,
        kHeadButt,
        kHealingRay,
        kHeatRay,
        kImpact,       /* football, improvised thrown weapon */
        kIncendiary,
        kKick,
        kLaser,
        kLight,        /* blinding flash */
        kLegalThreat,
        kLightningBolt,
        kMentalBlast,
        kPlasmaGlob,   /* plasma guns */
        kPoisonRay,
        kPrick,        /* Gom Jabbar needle */
        kPsionicStorm,
        kPunch,
        kQuill,
        kRail,         /* railgun bullet */
        kRestorationRay,
        kShot,         /* shotgun pellets */
        kSlash,
        kSlime,
        kSmash,
        kSpear,        /* speargun short spear */
        kSpit,         /* for hydralisk */
        kStab,
        kStasisRay,
        kSting,
        kSword,
        kTailSlap,
        kTouch,
        kTransporterRay,
        kVomit,
        kWeb,
        kWhip,
        kZap,
        kMaxAttackType
    };


    enum Effect {
        kSingle,    /* single target */
        kSmartDisc, /* for Yautja Smart-Discs */
        kCone,      /* a spreading area-effect attack */
        kExtend,    /* a straight line area-effect attack */
        kBeam,      /* as above but with random 1-6 bonus to range */
        kBurst,     /* a disc that doesn't penetrate walls */
        kFarBurst,  /* kBurst centered at target hit by kSingle */
        kOther
    };

    shObjId mWeaponIlk;      /* TODO: Scrap this unused variable. */
    shMutantPower mPower;
    Type mType;
    Effect mEffect;
    int mFlags;
    int mRadius;             /* blast radius in squares */
    int mRange;              /* in squares for beams (+d6 will be added),
                                o/w in feet */
    int mProb;               /* melee: rel. prob. of monst choosing this attack
                                ranged: inverse prob of using this attack */
    int mAttackTime;         /* AP spent recovering after attack */
    struct {
        shEnergyType mEnergy;
        int mLow, mHigh; /* damage ranges */
        int mChance;     /* percentage chance of causing this damage */
    } mDamage[3];

    shAttack () {}

    shAttack (shObjId weaponilk, Type type, Effect effect, int flags,
              int attacktime, int prob, int range, int radius,
              shEnergyType energytype0, int low0, int high0, int chance0,
              shEnergyType energytype1, int low1, int high1, int chance1,
              shEnergyType energytype2, int low2, int high2, int chance2)
    {
        mWeaponIlk = weaponilk;
        mType = type;
        mEffect = effect;
        mFlags = flags;
        mRange = range;
        mRadius = radius;
        mAttackTime = attacktime;
        mProb = prob;
        mDamage[0].mEnergy = energytype0;
        mDamage[0].mLow = low0;
        mDamage[0].mHigh = high0;
        mDamage[0].mChance = chance0;
        mDamage[1].mEnergy = energytype1;
        mDamage[1].mLow = low1;
        mDamage[1].mHigh = high1;
        mDamage[1].mChance = chance1;
        mDamage[2].mEnergy = energytype2;
        mDamage[2].mLow = low2;
        mDamage[2].mHigh = high2;
        mDamage[2].mChance = chance2;
    }

    int bypassesShield ();
    int isSingleAttack () { return kSingle == mEffect; }
    int isAimedAttack () { return kAimed & mFlags; }
    int isMeleeAttack () { return kMelee & mFlags; }
    int isMissileAttack () { return kMissile & mFlags; }
    int isTouchAttack () { return kTouchAttack & mFlags; }
    int isSpecialAttack () { return IS_SPECIAL_ENERGY (mDamage[0].mEnergy); }
    int isSecondaryAttack () { return kSecondaryAttack & mFlags; }
    int isAbsorbedForFree (int damage);
    int isLightGenerating (); /* drawn over dark squares */
    int dealsEnergyType (shEnergyType type);

    int getThreatRange (shObject *weapon, shCreature *target);
    int getCriticalMultiplier (shObject *weapon, shCreature *target);

    const char *noun ();
    const char *damStr ();
};

const char *energyDescription (shEnergyType t);

shAttack *newSingleAttack (shAttack::Type type, int flags,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1 = kNoEnergy, int low1 = 0, int high1 = 0,
    shEnergyType energy2 = kNoEnergy, int low2 = 0, int high2 = 0);
shAttack *newOtherAttack (shAttack::Type type, int flags,
    shEnergyType energy, int low, int high);
shAttack *newBurstAttack (shAttack::Type type, int flags, int radius,
    shEnergyType energy, int low, int high);
shAttack *newRayGunAttack (shAttack::Type type,
    shEnergyType energy, int low, int high);
shAttack *newMeleeAttack (shAttack::Type type, int time,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1 = kNoEnergy, int low1 = 0, int high1 = 0,
    shEnergyType energy2 = kNoEnergy, int low2 = 0, int high2 = 0);
shAttack *newMeleeAttackP (shAttack::Type type, int time,
    int prob0, shEnergyType energy0, int low0, int high0,
    int prob1 = 0, shEnergyType energy1 = kNoEnergy, int low1 = 0, int high1 = 0,
    int prob2 = 0, shEnergyType energy2 = kNoEnergy, int low2 = 0, int high2 = 0);
shAttack *newMissileAttack (shAttack::Type type, int time, int range,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1 = kNoEnergy, int low1 = 0, int high1 = 0,
    shEnergyType energy2 = kNoEnergy, int low2 = 0, int high2 = 0);
shAttack *newGrenadeAttack (shAttack::Type type, int radius,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1 = kNoEnergy, int low1 = 0, int high1 = 0,
    shEnergyType energy2 = kNoEnergy, int low2 = 0, int high2 = 0);
shAttack *newGunAttack (shAttack::Type type, int time, int range,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1 = kNoEnergy, int low1 = 0, int high1 = 0,
    shEnergyType energy2 = kNoEnergy, int low2 = 0, int high2 = 0);
shAttack *newGunAttackP (shAttack::Type type, int time, int range,
    int prob0, shEnergyType energy0, int low0, int high0,
    int prob1 = 0, shEnergyType energy1 = kNoEnergy, int low1 = 0, int high1 = 0,
    int prob2 = 0, shEnergyType energy2 = kNoEnergy, int low2 = 0, int high2 = 0);
shAttack *newBeamAttack (shAttack::Type type, int time, int range,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1 = kNoEnergy, int low1 = 0, int high1 = 0,
    shEnergyType energy2 = kNoEnergy, int low2 = 0, int high2 = 0);
shAttack *newBeamAttackP (shAttack::Type type, int time, int range,
    int prob0, shEnergyType energy0, int low0, int high0,
    int prob1 = 0, shEnergyType energy1 = kNoEnergy, int low1 = 0, int high1 = 0,
    int prob2 = 0, shEnergyType energy2 = kNoEnergy, int low2 = 0, int high2 = 0);
shAttack *newConeAttack (shAttack::Type type, int time, int range, int radius,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1 = kNoEnergy, int low1 = 0, int high1 = 0,
    shEnergyType energy2 = kNoEnergy, int low2 = 0, int high2 = 0);
shAttack *newMonsterAttack (shAttack::Type type, int flags, int time, int prob,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1 = kNoEnergy, int low1 = 0, int high1 = 0,
    shEnergyType energy2 = kNoEnergy, int low2 = 0, int high2 = 0);
shAttack *newMonsterRangedAttack (shAttack::Type type, shAttack::Effect effect,
    int flags, int time, int prob, int range,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1 = kNoEnergy, int low1 = 0, int high1 = 0,
    shEnergyType energy2 = kNoEnergy, int low2 = 0, int high2 = 0);

shAttack *newAttack (shAttack::Type type, shAttack::Effect effect,
    int flags, int time, int range, int radius,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1 = kNoEnergy, int low1 = 0, int high1 = 0,
    shEnergyType energy2 = kNoEnergy, int low2 = 0, int high2 = 0);

extern shAttack *ImprovisedObjectAttack;
extern shAttack *ImprovisedMissileAttack;
extern shAttack *KickedFootballAttack;
extern shAttack *GroundCollisionAttack;
extern shAttack *PitTrapDamage;
extern shAttack *AcidPitTrapDamage;
extern shAttack *AcidPitSipDamage;
extern shAttack *OpticBlastAttack;
extern shAttack *AcidBloodAttack;
extern shAttack *ExplodingMonsterAttack;
extern shAttack *FallingVatAttack;
extern shAttack *RestrainingBoltWear;
extern shAttack *TorcRemoveAttack;
extern shAttack *GloryDeviceAttack[5];

#endif
